# Copyright 2021-2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Main packer module."""

import argparse
import errno
import glob
import json
import logging
import os
import pathlib
import shutil
import tempfile
import uuid
import venv
import zipapp
from collections import namedtuple
from pathlib import Path
from typing import List

from pyempaq import __version__
from pyempaq.common import find_venv_bin, logged_exec, ExecutionError, PackError
from pyempaq.config_manager import load_config, ConfigError, Config


# setup logging
logging.basicConfig(
    format='%(asctime)s.%(msecs)03d %(levelname)-5s %(message)s', datefmt='%H:%M:%S')
logger = logging.getLogger(__name__)

# collected arguments
Args = namedtuple("Args", "project_name basedir entrypoint requirement_files")

# dependencies needed for the unpacker to run ok
UNPACKER_DEPS = ["packaging", "platformdirs"]


def get_pip():
    """Ensure an usable version of `pip`."""
    useful_pip = Path("pip3")
    # try to see if it's already installed
    try:
        logged_exec([useful_pip, "--version"])
    except ExecutionError:
        # failed to find a runnable pip, we need to install one
        pass
    else:
        return useful_pip

    # no useful pip found, let's create a virtualenv and use the one inside
    tmpdir = Path(tempfile.mkdtemp())
    venv.create(tmpdir, with_pip=True)
    useful_pip = find_venv_bin(tmpdir, "pip3")
    return useful_pip


def copy_project(src_dir: Path, dest_dir: Path, include: List[str], exclude: List[str]):
    """Copy/link the selected project content from the source to the destination directory.

    The content is selected with the 'include' list of patterns, minus the 'exclude' ones.

    It works differently for the different node types:
    - regular files: hard links (unless permission error or cross device)
    - directories: created
    - symlinks: respected, validating that they don't link to outside
    - other types (blocks, mount points, etc): ignored
    """
    included_nodes = {}  # use a dict because we want to avoid duplicates, but we care about order
    included_nodes["."] = None  # always the root, to create the destination directory
    excluded_nodes = set()

    # XXX Facundo 2023-04-07: this whole try/finally around changing directories can be
    # simplified to just pass `root_dir` to `glob.glob` when we stop supporting < 3.10
    _original_dir = os.getcwd()
    os.chdir(src_dir)
    try:
        for pattern in include:
            items = glob.glob(pattern, recursive=True)
            if items:
                included_nodes.update(dict.fromkeys(items))
            else:
                logger.error("Cannot find nodes for specified pattern: %r", pattern)

        # get all excluded nodes, building the source path so it's
        # easier comparable in the main loop below
        for pattern in exclude:
            excluded_nodes.update(src_dir / path for path in glob.iglob(pattern, recursive=True))
    finally:
        os.chdir(_original_dir)

    # need to remove all content inside symlinked directories (as that symlink will
    # be reproduced, so those contents don't need to be particularly handled)
    symlinked_dirs = set()
    for node in included_nodes:
        node = src_dir / node
        if node.is_dir() and node.is_symlink():
            symlinked_dirs.add(node)

    def _relative(node):
        """Return str'ed node relative to src_dir, ready to log."""
        return str(node.relative_to(src_dir))

    for node in included_nodes:
        src_node = src_dir / node
        dest_node = dest_dir / node

        if src_node in excluded_nodes:
            logger.debug("Ignoring excluded node: %r", _relative(src_node))
            continue
        if any(parent in excluded_nodes for parent in src_node.parents):
            logger.debug("Ignoring node because excluded parent: %r", _relative(src_node))
            continue
        if any(parent in symlinked_dirs for parent in src_node.parents):
            # node is inside a symlinked path, no need to duplicate it
            continue

        # if included node is only part of subtree, ensure parent directories are there
        if not dest_node.parent.exists():
            dest_node.parent.mkdir(parents=True)

        if src_node.is_symlink():
            real_pointed_node = src_node.resolve()
            if src_dir not in real_pointed_node.parents:
                logger.debug(
                    "Ignoring symlink because targets outside the project: %r -> %r",
                    _relative(src_node), str(real_pointed_node),
                )
                continue
            relative_link = os.path.relpath(real_pointed_node, src_node.parent)
            target_is_dir = real_pointed_node.is_dir()  # needed for Windows
            dest_node.symlink_to(relative_link, target_is_directory=target_is_dir)

        elif src_node.is_dir():
            dest_node.mkdir(mode=src_node.stat().st_mode, exist_ok=True)

        elif src_node.is_file():
            try:
                # XXX Facundo 2023-04-07: we can use simpler `dest_node.hardlink_to(src_node)`
                # when we stop supporting < 3.10
                os.link(src_node, dest_node)
            except OSError as error:
                if error.errno != errno.EXDEV and not isinstance(error, PermissionError):
                    raise
                shutil.copy2(src_node, dest_node)

        else:
            logger.debug("Ignoring file because of type: %r", _relative(src_node))


def prepare_metadata(origdir: pathlib.Path, config: Config):
    """Prepare the meta-data for the future unpacker action.

    Note that all paths in the config are all already validated to exist
    and relative to the base directory (so no "place adaptation" needs
    to happen for the unpacking).
    """
    # store the needed metadata
    logger.debug("Saving metadata from config %s", config)
    metadata = {
        "requirement_files": [str(path) for path in config.requirements],
        "project_name": config.name,
        "exec_default_args": config.exec.default_args,
        "unpack_restrictions": dict(config.unpack_restrictions or {}),
    }
    if config.exec.script is not None:
        metadata["exec_style"] = "script"
        metadata["exec_value"] = str(config.exec.script)
    elif config.exec.module is not None:
        metadata["exec_style"] = "module"
        metadata["exec_value"] = str(config.exec.module)
    elif config.exec.entrypoint is not None:
        metadata["exec_style"] = "entrypoint"
        metadata["exec_value"] = config.exec.entrypoint

    # if dependencies, store them just as another requirement file (save it inside the project,
    # but using an unique name to not overwrite anything)
    if config.dependencies:
        unique_name = f"pyempaq-autoreq-{uuid.uuid4()}.txt"
        extra_deps = origdir / unique_name
        extra_deps.write_text("\n".join(config.dependencies) + "\n")
        metadata["requirement_files"].append(unique_name)

    return metadata


def pack(config):
    """Pack."""
    pyempaq_source_root = Path(__file__).parent
    tmpdir = Path(tempfile.mkdtemp())
    logger.debug("Working in temp dir %r", str(tmpdir))

    # copy all the project content inside "orig" in temp dir
    origdir = tmpdir / "orig"
    copy_project(config.basedir, origdir, config.include, config.exclude)

    # ensure all requirements are included by users
    missing_requirements = []
    for path in config.requirements:
        if not (origdir / path).exists():
            missing_requirements.append(str(path))

    if missing_requirements:
        raise PackError(
            f"The indicated requirements {missing_requirements} are not included "
            "along the packed files; ensure to include them explicitly in the config."
        )

    # copy the common module
    pyempaq_dir = tmpdir / "pyempaq"
    pyempaq_dir.mkdir()
    common_src = pyempaq_source_root / "common.py"
    common_final_src = tmpdir / "pyempaq" / "common.py"
    shutil.copy(common_src, common_final_src)

    # copy the unpacker as the entry point of the zip
    unpacker_src = pyempaq_source_root / "unpacker.py"
    unpacker_final_main = tmpdir / "__main__.py"
    shutil.copy(unpacker_src, unpacker_final_main)

    # build a dir with the dependencies needed by the unpacker
    logger.debug("Building internal dependencies dir")
    venv_dir = tmpdir / "venv"
    pip = get_pip()
    cmd = [pip, "install", *UNPACKER_DEPS, f"--target={venv_dir}"]
    logged_exec(cmd)

    metadata = prepare_metadata(origdir, config)
    metadata_file = tmpdir / "metadata.json"
    with metadata_file.open("wt", encoding="utf8") as fh:
        json.dump(metadata, fh)

    # create the zipfile
    packed_filepath = f"{config.name}.pyz"
    zipapp.create_archive(tmpdir, packed_filepath)

    # clean the temporary directory
    shutil.rmtree(tmpdir)

    logger.info("Done, project packed in %r", str(packed_filepath))


def main():
    """Manage CLI interaction and call pack."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "source", type=Path,
        help="The source file (pyempaq.yaml) or the directory where to find it.")
    parser.add_argument(
        '-v', '--verbose',
        help="Show detailed information, typically of interest only when diagnosing problems.",
        action="store_const", dest="loglevel", const=logging.DEBUG)
    parser.add_argument(
        '-q', '--quiet',
        help="Only events of WARNING level and above will be tracked.",
        action="store_const", dest="loglevel", const=logging.WARNING)
    parser.add_argument(
        '-V', '--version',
        help="Print the version and exit.",
        action="version", version=__version__)
    args = parser.parse_args()

    if args.loglevel is not None:
        logging.getLogger().setLevel(args.loglevel)

    try:
        logger.info("Parsing configuration in %r", str(args.source))
        config = load_config(args.source)
    except ConfigError as err:
        logger.error(err)
        for err in err.errors:
            logger.error(err)
        exit(1)

    try:
        pack(config)
    except PackError as exc:
        logger.error("Pack error: %s", exc)
        exit(2)
    except Exception as exc:
        logger.error("Unknown internal error: %r", exc)
        exit(3)



# Copyright 2022 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Common functionality for packer and unpucker modules."""

import logging
import subprocess


logger = logging.getLogger('logger')


class ExecutionError(Exception):
    """The subprocess didn't finish ok."""


class PackError(Exception):
    """An error occurred while packing the project."""


def find_venv_bin(basedir, exec_base):
    """Heuristics to find the pip executable in different platforms."""
    bin_dir = basedir / "bin"
    if bin_dir.exists():
        # linux-like environment
        return bin_dir / exec_base

    bin_dir = basedir / "Scripts"
    if bin_dir.exists():
        # windows environment
        return bin_dir / f"{exec_base}.exe"

    raise RuntimeError(f"Binary not found inside venv; subdirs: {list(basedir.iterdir())}")


def logged_exec(cmd):
    """Execute a command, redirecting the output to the log."""
    cmd = list(map(str, cmd))
    logger.debug(f"Executing external command: {cmd}")
    try:
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except Exception as err:
        raise ExecutionError(f"Command {cmd} crashed with {err!r}")
    stdout = []
    for line in proc.stdout:
        line = line[:-1]
        stdout.append(line)
        logger.debug(f":: {line}")
    retcode = proc.wait()
    if retcode:
        raise ExecutionError(f"Command {cmd} ended with retcode {retcode}")
    return stdout



# Copyright 2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Expose the version information at module level, as is standard."""

from pyempaq._version import __version__, VERSION  # noqa



# Copyright 2021-2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""The configuration manager."""

import pathlib
from typing import List, Optional

import pydantic
import yaml
from typing_extensions import Annotated


# base directory to which the different included paths may be relative from
_BASEDIR = None

# directory where the config is taken from, which is the default for basedir
_CONFIGDIR = None

# the default include value to get all the project inside
DEFAULT_INCLUDE_LIST = ["./**"]


class ConfigError(Exception):
    """Specific errors found in the config."""

    def __init__(self, msg=None, errors=None):
        if msg is None:
            msg = "Problem(s) found in configuration file:"
        super().__init__(msg)
        if errors is None:
            errors = []
        self.errors = errors


def _relative_path_validator(value):
    """Constrained string which must be a relative path."""
    if value is None:
        return
    value = pathlib.Path(value)

    if value.is_absolute():
        raise AssertionError("path must be relative")

    # relative to the basedir
    abs_path = _BASEDIR / value

    if not abs_path.exists():
        raise AssertionError(f"path {str(abs_path)!r} not found")

    if _BASEDIR not in abs_path.resolve().parents:
        raise AssertionError("relative path must be inside the packed project")

    # return the relative path
    return value


def _relative_file_validator(value):
    """Constrained relative path which must be a file."""
    value = _relative_path_validator(value)
    if value is None:
        return

    # relative to the basedir
    abs_path = _BASEDIR / value

    if not abs_path.is_file():
        raise AssertionError(f"path {str(abs_path)!r} must be a file")

    # return the relative path
    return value


RelativePath = Annotated[str, pydantic.AfterValidator(_relative_path_validator)]
RelativeFile = Annotated[str, pydantic.AfterValidator(_relative_file_validator)]


class ModelConfigDefaults(pydantic.BaseModel):
    """Define defaults for the BaseModel configuration."""

    model_config = dict(
        extra="forbid",
        frozen=True,
        alias_generator=lambda s: s.replace("_", "-"),
    )


class Executor(ModelConfigDefaults, alias_generator=lambda s: s.replace("_", "-")):
    """Executor information."""

    script: Optional[RelativeFile] = None
    module: Optional[RelativePath] = None
    entrypoint: Optional[List[pydantic.StrictStr]] = None
    default_args: List[pydantic.StrictStr] = []


class UnpackRestrictions(ModelConfigDefaults, alias_generator=lambda s: s.replace("_", "-")):
    """Restrictions that will be verified/enforced during unpack."""

    minimum_python_version: pydantic.StrictStr = None


class Config(ModelConfigDefaults):
    """Definition of PyEmpaq's configuration."""

    name: str
    basedir: pathlib.Path  # the default for this is managed on YAML load time
    exec: Executor
    requirements: List[RelativeFile] = []
    dependencies: List[str] = []
    include: List[str] = DEFAULT_INCLUDE_LIST
    exclude: List[str] = []
    unpack_restrictions: Optional[UnpackRestrictions] = None

    @pydantic.field_validator("basedir")
    def ensure_basedir(cls, value):
        """Ensure that the basedir is valid, and store it to be used by other paths."""
        global _BASEDIR

        value = value.expanduser()
        if not value.is_absolute():
            value = _CONFIGDIR / value

        # set the basedir after the expanding/absolutizing, so the rest of the config can use it
        _BASEDIR = value

        if not value.exists():
            raise AssertionError(f"path {str(value)!r} not found")
        if not value.is_dir():
            raise AssertionError(f"path {str(value)!r} must be a directory")
        return value

    @pydantic.field_validator("exec", mode="before")
    def validate_exec_subkeys(cls, values):
        """Check the exec subkeys."""
        # it must be one, and only one, of these...
        subkeys = ["script", "module", "entrypoint"]
        count = sum(key in values for key in subkeys)
        if count == 0:
            subkeys_str = ', '.join(repr(x) for x in subkeys)
            raise AssertionError(f"need at least one of these subkeys: {subkeys_str}")
        if count > 1:
            subkeys_str = ', '.join(repr(x) for x in subkeys)
            raise AssertionError(f"only one of these subkeys is allowed: {subkeys_str}")
        return values


def _format_pydantic_errors(errors):
    """Format pydantic errors for a simpler presentation."""
    formated_errors = []
    for error in errors:
        # format location
        loc_parts = []
        for part in error['loc']:
            if isinstance(part, int):
                # an index, fix previous part
                loc_parts[-1] = f"{loc_parts[-1]}[{part}]"
            else:
                loc_parts.append(str(part))
        location = ".".join(loc_parts)

        message = error['msg'].strip().removeprefix("Assertion failed, ")
        formated_errors.append(f"- {location!r}: {message}")

    return formated_errors


def load_config(path):
    """Load the config from charmcraft.yaml in the indicated directory."""
    global _CONFIGDIR

    path = path.expanduser().absolute()
    if path.is_dir():
        _CONFIGDIR = path
        configpath = path / "pyempaq.yaml"
    else:
        _CONFIGDIR = path.parent
        configpath = path

    if not configpath.exists():
        raise ConfigError(f"Configuration file not found: {str(configpath)!r}")

    try:
        content = yaml.safe_load(configpath.read_text())
    except Exception:
        raise ConfigError(f"Cannot open and parse YAML configuration file {str(configpath)!r}")

    # if `basedir` defaults to the directory where the configuration exists
    if "basedir" not in content:
        content["basedir"] = configpath.parent

    try:
        parsed = Config.model_validate(content)
    except pydantic.ValidationError as error:
        raise ConfigError(errors=_format_pydantic_errors(error.errors()))

    return parsed



# Copyright 2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Holder of the PyEmpaq version number."""

# these two will be exported at `pyempaq` module level by __init__.py; also VERSION will
# be parsed by setup.py without needing to import the module (at moments when the
# version is needed but the infrastructure is not in place yet)
VERSION = (0, 6, 0)
__version__ = '.'.join([str(x) for x in VERSION])



# Copyright 2021-2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Unpacking functionality.."""

import enum
import hashlib
import importlib
import json
import logging
import os
import pathlib
import platform
import shutil
import subprocess
import sys
import time
import venv
import zipfile
from types import ModuleType
from typing import List, Dict, Any

from pyempaq.common import find_venv_bin, logged_exec


# this is the directory for the NEW virtualenv created for the project (not the packed
# one to run unpacker itself)
PROJECT_VENV_DIR = "project_venv"

# the file name to flag that the project setup completed successfully
COMPLETE_FLAG_FILE = "complete.flag"

# the real magic number is a byte sequence with "\r\n" at the end; it's built here
# so it's easily patchable by tests
MAGIC_NUMBER = importlib.util.MAGIC_NUMBER[:-2].hex()

# the environment variable to specify a different action
ACTION_ENVVAR = "PYEMPAQ_ACTION"

# setup logging
logger = logging.getLogger()
handler = logging.StreamHandler()
fmt = logging.Formatter("::pyempaq:: %(asctime)s %(message)s")
handler.setFormatter(fmt)
handler.setLevel(0)
logger.addHandler(handler)
logger.setLevel(logging.ERROR if os.environ.get("PYEMPAQ_DEBUG") is None else logging.INFO)


class FatalError(Exception):
    """Error that will terminate the unpacking.

    This is never shown to the user, but the overall process will exit with
    the indicated return code.
    """

    class ReturnCode(enum.IntEnum):
        """Codes that the unpacker may return."""

        restrictions_not_met = 64
        unpack_basedir_missing = 65
        unpack_basedir_notdir = 66
        bad_action = 67

    def __init__(self, code):
        self.returncode = code
        super().__init__("")


# -- special PyEmpaq actions

def special_action_info(pyempaq_dir, metadata):
    """Provide information about different installations for the project."""
    print("Base PyEmpaq directory:", pyempaq_dir)
    subdirs = sorted(pyempaq_dir.glob(f"{metadata['project_name']}-*"))
    if not subdirs:
        print("No installation found!")
        return

    print("Current installations:")
    for subdir in subdirs:
        print("   ", subdir.name)


def special_action_uninstall(pyempaq_dir, metadata):
    """Remove all installations for the given project."""
    subdirs = sorted(pyempaq_dir.glob(f"{metadata['project_name']}-*"))
    if not subdirs:
        print("No installation found!")
        return

    print("Removing installation:")
    for subdir in subdirs:
        print("   ", subdir.name)
        shutil.rmtree(subdir)


ACTION_CHOICES = {
    "info": special_action_info,
    "uninstall": special_action_uninstall,
}

# --


def get_python_exec(project_dir: pathlib.Path) -> pathlib.Path:
    """Return the Python exec to use.

    If a venv is present (just created or from a previous unpack) use it, else just use
    the one used to run this script.
    """
    venv_dir = project_dir / PROJECT_VENV_DIR
    if venv_dir.exists():
        executable = find_venv_bin(venv_dir, "python")
    else:
        executable = pathlib.Path(sys.executable)
    return executable


def build_command(python_exec: str, metadata: Dict[str, str], sys_args: List[str]) -> List[str]:
    """Build the command to be executed."""
    if metadata["exec_style"] == "script":
        cmd = [python_exec, metadata["exec_value"]]
    elif metadata["exec_style"] == "module":
        cmd = [python_exec, "-m", metadata["exec_value"]]
    elif metadata["exec_style"] == "entrypoint":
        cmd = [python_exec] + metadata["exec_value"]

    if sys_args:
        cmd.extend(sys_args)
    else:
        cmd.extend(metadata["exec_default_args"])
    return cmd


def run_command(venv_bin_dir: pathlib.Path, cmd: List[str]) -> subprocess.CompletedProcess:
    """Run the command with a custom context."""
    newenv = os.environ.copy()
    venv_bin_dir_str = str(venv_bin_dir)
    if "PATH" in newenv:
        newenv["PATH"] = newenv["PATH"] + ":" + venv_bin_dir_str
    else:
        newenv["PATH"] = venv_bin_dir_str
    newenv["PYEMPAQ_PYZ_PATH"] = os.path.dirname(__file__)
    return subprocess.run(cmd, env=newenv)


def get_file_hexdigest(filepath: pathlib.Path) -> str:
    """Hash a file and return its hexdigest."""
    hasher = hashlib.sha256()
    with open(filepath, "rb") as fh:
        while True:
            data = fh.read(65536)
            hasher.update(data)
            if not data:
                break
    return hasher.hexdigest()


def setup_project_directory(
    zf: zipfile.ZipFile,
    project_dir: pathlib.Path,
    venv_requirements: List[pathlib.Path],
    *,
    ephemeral=False,
):
    """Set up the project directory (if needed).

    Process in case it's needed:

    - create the directory

    - extract everything from the zipfile into the new directory

    - if there are virtualenv dependencies:

        - create the virtualenv

        - install dependencies there

    After successful set up a flag is left in the directory so next time the unpacker is run
    it recognizes everything is done (note that having the directory is not enough, it may
    have been partially set up).
    """
    if project_dir.exists():
        if (project_dir / COMPLETE_FLAG_FILE).exists():
            log_call = logger.warning if ephemeral else logger.info
            log_call("Reusing project dir %r", str(project_dir))
            return
        logger.info("Found incomplete project dir %r", str(project_dir))
        shutil.rmtree(project_dir)
        logger.info("Removed old incomplete dir")

    logger.info("Creating project dir %r", str(project_dir))
    project_dir.mkdir()

    logger.info("Extracting pyempaq content")
    zf.extractall(path=project_dir)

    if venv_requirements:
        logger.info("Creating payload virtualenv")
        venv_dir = project_dir / PROJECT_VENV_DIR
        venv.create(venv_dir, with_pip=True)
        pip_exec = find_venv_bin(venv_dir, "pip3")
        cmd = [str(pip_exec), "install"]
        for req_file in venv_requirements:
            cmd += ["-r", str(req_file)]
        logger.info("Installing dependencies: %s", cmd)
        logged_exec(cmd)
        logger.info("Virtualenv setup finished")
    else:
        logger.info("Skipping virtualenv (no requirements)")

    # store unpacking metadata
    zf_hash = get_file_hexdigest(zf.filename)
    metadata = {
        "pyz_path": str(zf.filename),
        "pyz_hash": zf_hash,
        "timestamp": time.time(),
    }
    (project_dir / "unpacking.json").write_text(json.dumps(metadata))

    # save the flag for completeness
    (project_dir / COMPLETE_FLAG_FILE).touch()


def enforce_restrictions(version: ModuleType, restrictions: Dict[str, Any]) -> bool:
    """Enforce the unpacking restrictions, if any; raise a fatal error if they are not met."""
    if not restrictions:
        return
    ignored_restrictions = os.environ.get("PYEMPAQ_IGNORE_RESTRICTIONS", "").split(",")

    mpv = restrictions.get("minimum_python_version")
    if mpv is not None:
        current = platform.python_version()
        logger.info("Checking minimum Python version: indicated=%r current=%r", mpv, current)
        if version.parse(mpv) > version.parse(current):
            msg = "Failed to comply with version restriction: need at least Python %s"
            if "minimum-python-version" in ignored_restrictions:
                logger.info("(ignored) " + msg, mpv)
            else:
                logger.error(msg, mpv)
                raise FatalError(FatalError.ReturnCode.restrictions_not_met)


def build_project_install_dir(zip_path: pathlib.Path, metadata: Dict[str, str]):
    """Build the name of the directory where everything will be extracted."""
    project_name = metadata["project_name"]

    # get the first part of the hash of the file
    hexdigest = get_file_hexdigest(zip_path)
    file_hash_partial = hexdigest[:20]

    # Python details
    py_impl = platform.python_implementation().lower()
    py_version = ".".join(platform.python_version_tuple()[:2])

    name = f"{project_name}-{file_hash_partial}-{py_impl}.{py_version}.{MAGIC_NUMBER}"
    return name


def get_base_dir(platformdirs: ModuleType) -> pathlib.Path:
    """Get the base directory for all PyEmpaq installs.

    If it's indicated by the user, it must exist and be a directory (less error prone
    this way). If it's the default, this function ensures it is created.

    It needs to receive the 'platformdirs' because it's imported from the builtin
    virtualenv (not generically available).
    """
    custom_basedir = os.environ.get("PYEMPAQ_UNPACK_BASE_PATH")
    if custom_basedir is None:
        basedir = pathlib.Path(platformdirs.user_data_dir()) / 'pyempaq'
        basedir.mkdir(parents=True, exist_ok=True)
    else:
        basedir = pathlib.Path(custom_basedir)
        if not basedir.exists():
            raise FatalError(FatalError.ReturnCode.unpack_basedir_missing)
        if not basedir.is_dir():
            raise FatalError(FatalError.ReturnCode.unpack_basedir_notdir)

    return basedir


def run():
    """Run the unpacker."""
    logger.info("PyEmpaq start")

    # parse pyempaq metadata from the zip file
    pyempaq_filepath = pathlib.Path.cwd() / sys.argv[0]
    zf = zipfile.ZipFile(pyempaq_filepath)
    metadata = json.loads(zf.read("metadata.json").decode("utf8"))
    logger.info("Loaded metadata: %s", metadata)

    # load platformdirs and packaging from the builtin venv (not at top of file because
    # paths needed to be fixed)
    sys.path.insert(0, f"{pyempaq_filepath}/venv/")
    import platformdirs  # NOQA
    from packaging import version  # NOQA

    # check all restrictions are met
    enforce_restrictions(version, metadata["unpack_restrictions"])

    pyempaq_dir = get_base_dir(platformdirs)
    logger.info("Base directory: %r", str(pyempaq_dir))

    indicated_action = os.environ.get(ACTION_ENVVAR)
    if indicated_action is not None:
        indicated_action = indicated_action.lower()
        print(f"Running {indicated_action!r} action")
        try:
            func = ACTION_CHOICES[indicated_action]
        except KeyError:
            logger.error(
                "Bad specified action %s=%r; valid options: %s",
                ACTION_ENVVAR, indicated_action, ", ".join(ACTION_CHOICES.keys()))
            raise FatalError(FatalError.ReturnCode.bad_action)
        return func(pyempaq_dir, metadata)

    # create a temp dir and extract the project there
    project_dir = pyempaq_dir / build_project_install_dir(pyempaq_filepath, metadata)
    original_project_dir = project_dir / "orig"
    venv_requirements = [original_project_dir / fname for fname in metadata["requirement_files"]]
    ephemeral = os.environ.get("PYEMPAQ_EPHEMERAL")
    setup_project_directory(zf, project_dir, venv_requirements, ephemeral=ephemeral)

    python_exec = get_python_exec(project_dir)
    original_process_directory = os.getcwd()
    os.chdir(original_project_dir)

    cmd = build_command(str(python_exec), metadata, sys.argv[1:])
    logger.info("Running payload: %s", cmd)
    venv_bin_dir = python_exec.parent
    proc = run_command(venv_bin_dir, cmd)
    logger.info("Exit code: %s", proc.returncode)

    if ephemeral:
        logger.info("Removing project install directory because ephemeral indicated.")
        os.chdir(original_process_directory)
        shutil.rmtree(project_dir)
    logger.info("PyEmpaq done")
    return proc.returncode


if __name__ == "__main__":
    try:
        rc = run()
    except FatalError as error:
        rc = error.returncode
    exit(rc)
